#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Submodule for QDelta coefficients based on algebraic approaches,
in particular the `LU trick` from `[Weiser, 2014]`_.

Examples
--------
>>> from qmat.qcoeff.collocation import Collocation
>>> coll = Collocation(nNodes=4, nodeType="LEGENDRE", quadType="RADAU-RIGHT")
>>>
>>> from qmat import genQDeltaCoeffs
>>> QDelta = genQDeltaCoeffs("LU", Q=coll.Q)
>>>
>>> from qmat.qdelta.algebraic import LDU
>>> gen = LDU(Q=coll.Q)
>>> QDelta = gen.getQDelta()
"""
import numpy as np
import scipy.linalg as spl

from qmat.qdelta import QDeltaGenerator, register
from qmat.mathutils import lduFactorization


@register
class PIC(QDeltaGenerator):
    """Picard approximation (zeros)"""
    aliases = ["Picard"]

    def computeQDelta(self, k=None):
        return self.zeros


@register
class Exact(QDeltaGenerator):
    """Takes Q (exact approximation)"""
    aliases = ["EXACT"]

    def computeQDelta(self, k=None):
        return self.Q.copy()  # TODO: do we really want a copy here ... ?


@register
class LU(QDeltaGenerator):
    """LU approximation from `[Weiser, 2014] <https://link.springer.com/article/10.1007/s10543-014-0540-y>`_"""

    def computeQDelta(self, k=None):
        _, _, U = spl.lu(self.Q.T)
        return U.T


@register
class LU2(LU):
    """LU approximation from `[Weiser, 2014]`_ multiplied by 2"""

    def computeQDelta(self, k=None):
        return super().computeQDelta()*2


@register
class QPar(QDeltaGenerator):
    """Approximation using diagonal of Q"""
    aliases = ["Qpar", "Qdiag"]

    def computeQDelta(self, k=None):
        return np.diag(np.diag(self.Q))


@register
class GS(QDeltaGenerator):
    """Approximation using lower triangular part of Q"""
    aliases = ["GaussSeidel"]

    def computeQDelta(self, k=None):
        return np.tril(self.Q)


@register
class LDU(QDeltaGenerator):
    """Diagonal approximation using LDU factorization"""

    def computeQDelta(self, k=None):
        return lduFactorization(self.Q)[1]



import pytest
import numpy as np

import qmat.mathutils as mu

nNodeTests = [2, 3, 4, 5, 6, 7, 8]

@pytest.mark.parametrize("nNodes", nNodeTests)
def testRegression(nNodes):

    nodes = np.linspace(0, 1, num=nNodes, endpoint=False)

    for pOrder in range(1, nNodes):
        times = nodes + 1
        Pe = mu.getExtrapolationMatrix(nodes, times, pOrder)

        polyCoeffs = np.random.rand(pOrder+1)
        nodeValues = np.polyval(polyCoeffs, nodes)
        refValues = np.polyval(polyCoeffs, times)

        assert np.allclose(refValues, Pe @ nodeValues)


@pytest.mark.parametrize("nNodes", nNodeTests)
def testLDUFactorization(nNodes):
    Q = np.random.rand(nNodes, nNodes)
    L, D, U = mu.lduFactorization(Q)
    assert np.allclose(Q, L @ D @ U)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
QDelta coefficients based on minimization approaches.
In particular, generates the diagonal coefficients from `[Caklovic et al., 2024]`_.

Examples
--------
>>> from qmat.qcoeff.collocation import Collocation
>>> coll = Collocation(nNodes=4, nodeType="LEGENDRE", quadType="RADAU-RIGHT")
>>>
>>> from qmat import genQDeltaCoeffs
>>> QDelta = genQDeltaCoeffs("MIN-SR-NS", nodes=coll.nodes)
>>>
>>> from qmat.qdelta.min import MIN_SR_S, MIN_SR_FLEX
>>> minSRS= MIN_SR_S(coll.nNodes, coll.nodeType, coll.quadType)
>>> QDelta = minSRS.getQDelta()
>>> minSRFLEX = MIN_SR_FLEX(coll.nNodes, coll.nodeType, coll.quadType)
>>> QD1, QD2, QD3 = minSRFLEX.genCoeffs(k=[1,2,3])
"""
import warnings
import numpy as np
import scipy.optimize as spo

from qmat.qdelta import QDeltaGenerator, register
import qmat.qdelta.mincoeffs as tables
from qmat.qcoeff.collocation import Collocation


def check(k):
    """Utility function to check k parameter for k-dependent generators"""
    if k is None: k = 1
    if k < 1:
        raise ValueError(f"k must be greater than 0 ({k})")
    return k


@register
class MIN(QDeltaGenerator):
    """Naive diagonal coefficients based on spectral radius optimization."""
    aliases = ["MIN-Speck"]

    def rho(self, x):
        M = self.size
        return max(abs(
            np.linalg.eigvals(np.eye(M) - np.diag(x).dot(self.Q))))

    def computeQDelta(self, k=None):
        x0 = 10 * np.ones(self.size)
        d = spo.minimize(self.rho, x0, method='Nelder-Mead')
        return np.linalg.inv(np.diag(d.x))


class FromTable(QDeltaGenerator):
    """Base (unregistered) class for diagonal coefficients stored in tables."""

    def __init__(self, nNodes, nodeType, quadType, **kwargs):
        self.nNodes = nNodes
        self.nodeType = nodeType
        self.quadType = quadType

    @property
    def size(self):
        return self.nNodes

    def computeQDelta(self, k=None):
        name = self.__class__.__name__
        try:
            table = getattr(tables, name)
            coeffs = table[self.nodeType][self.quadType][self.size]
            coeffs = np.asarray(coeffs, dtype=float)
            assert coeffs.ndim == 1
            assert coeffs.size == self.size
        except KeyError:
            raise NotImplementedError(
                f"no {name} MIN coefficients for"
                f"{self.nNodes} {self.nodeType} {self.quadType} nodes")
        except AssertionError:
            raise ValueError(
                f"MIN coefficients stored for {name} are inconsistent : {coeffs}")
        except:
            raise ValueError(
                f"could not convert {name} MIN coefficients to numpy array")
        return np.diag(coeffs)

    def check(cls):
        try:
            getattr(tables, cls.__name__)
        except AttributeError:
            raise AttributeError(
                f"no MIN coefficients table found for {cls.__name__}"
                " in qmat.qdelta.mincoeffs")
        return cls

def registerTable(cls:FromTable)->FromTable:
    return register(FromTable.check(cls))


@registerTable
class MIN3(FromTable):
    """Magic diagonal coefficients from `[Speck, 2021] <https://zenodo.org/records/5775971>`_."""
    aliases = ["Magic_Numbers"]


@registerTable
class MIN_VDHS(FromTable):
    """Diagonal coefficients from `[van der Houwen & Sommeijer, 1991] <https://epubs.siam.org/doi/10.1137/0912054>`_."""
    aliases = ["VDHS"]


@register
class MIN_SR_NS(QDeltaGenerator):
    """Diagonal `MIN-SR-NS` coefficients from `[Caklovic et al., 2024] <https://arxiv.org/pdf/2403.18641>`_."""

    aliases = ["MIN-SR-NS", "MIN_GT"]

    def __init__(self, nodes, **kwargs):
        self.nodes = np.asarray(nodes)

    @property
    def size(self):
        return self.nodes.size

    def computeQDelta(self, k=None):
        return np.diag(self.nodes)/self.size


@register
class MIN_SR_S(QDeltaGenerator):
    """Diagonal `MIN-SR-S` coefficients from `[Caklovic et al., 2024]`_."""

    aliases = ["MIN-SR-S"]

    def __init__(self, nNodes=None, nodeType=None, quadType=None, coll:Collocation=None, **kwargs):
        """
        Parameters
        ----------
        nNodes : int
            Number of nodes.
        nodeType : str
            Type of node distribution, see :class:`qmat.nodes.NodesGenerator`
            for available types.
        quadType : str
            Quadrature type for the nodes, see :class:`qmat.nodes.NodesGenerator`
            for available types.
        coll : :class:`qmat.qcoeff.collocation.Collocation`, optional
            If given, ignore the previous parameters and use this as the
            underlying collocation method. The default is None.
        **kwargs :
            Additional parameters given during a generic call, not used by this class.
        """
        if coll is not None:
            assert isinstance(coll, Collocation), "coll parameter is not a Collocation object"
            self.coll = coll
            self.nodeType = coll.nodeType
            self.quadType = coll.quadType
        else:
            self.nodeType = nodeType
            self.quadType = quadType
            self.coll = Collocation(nNodes, nodeType, quadType)

    @property
    def size(self):
        return self.coll.nodes.size

    def computeCoeffs(self, M, a=None, b=None):
        """
        Compute diagonal coefficients for a given number of nodes M.
        If `a` and `b` are given, then it uses as initial guess:

        >>> a * nodes**b / M

        If `a` is not given, then do not care about `b` and uses as initial guess:

        >>> nodes / M

        Parameters
        ----------
        M : int
            Number of collocation nodes.
        a : float, optional
            `a` coefficient for the initial guess.
        b : float, optional
            `b` coefficient for the initial guess.

        Returns
        -------
        coeffs : array
            The diagonal coefficients.
        nodes : array
            The nodes associated to the current coefficients.
        """
        nodeType, quadType = self.nodeType, self.quadType
        if M == self.size:
            collM = self.coll
        else:
            collM = Collocation(nNodes=M, nodeType=nodeType, quadType=quadType)
        QM, nodesM = collM.Q, collM.nodes

        if quadType in ['LOBATTO', 'RADAU-LEFT']:
            QM = QM[1:, 1:]
            nodesM = nodesM[1:]
        nCoeffs = len(nodesM)

        if nCoeffs == 1:
            coeffs = np.diag(QM)

        else:

            def nilpotency(coeffs):
                """Function verifying the nilpotency from given coefficients"""
                coeffs = np.asarray(coeffs)
                kMats = [(1 - z) * np.eye(nCoeffs) + z * np.diag(1 / coeffs) @ QM for z in nodesM]
                vals = [np.linalg.det(K) - 1 for K in kMats]
                return np.array(vals)

            if a is None:
                coeffs0 = nodesM / M
            else:
                coeffs0 = a * nodesM**b / M

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                coeffs = spo.fsolve(nilpotency, coeffs0, xtol=1e-15)

        # Handle first node equal to zero
        if quadType in ['LOBATTO', 'RADAU-LEFT']:
            coeffs = np.asarray([0.0] + list(coeffs))
            nodesM = np.asarray([0.0] + list(nodesM))

        return coeffs, nodesM

    @staticmethod
    def fit(coeffs, nodes):
        """Function fitting given coefficients to a power law"""

        def lawDiff(ab):
            a, b = ab
            return np.linalg.norm(a * nodes**b - coeffs)

        sol = spo.minimize(lawDiff, [1.0, 1.0], method="nelder-mead")
        return sol.x

    def computeQDelta(self, k=None):
        # Compute coefficients incrementally
        a, b = None, None
        m0 = 2 if self.quadType in ['LOBATTO', 'RADAU-LEFT'] else 1
        for m in range(m0, self.size + 1):
            coeffs, nodes = self.computeCoeffs(m, a, b)
            if m > 1:
                a, b = self.fit(coeffs * m, nodes)
        return np.diag(coeffs)


@register
class MIN_SR_FLEX(MIN_SR_S):
    """Diagonal `MIN-SR-FLEX` coefficients from `[Caklovic et al., 2024]`_"""

    aliases = ["MIN-SR-FLEX"]

    def computeQDelta(self, k=1):
        k = check(k)
        if k <= self.size:
            return np.diag(self.coll.nodes/k)
        else:
            try:
                self._QDelta_MIN_SR_S
            except AttributeError:
                self._QDelta_MIN_SR_S = super().computeQDelta()
            return self._QDelta_MIN_SR_S


@register
class Jumper(MIN_SR_NS):
    """Diagonal coefficients allowing order jump"""

    aliases = ["JUMPER", "FB"]

    def computeQDelta(self, k=1):
        k = check(k)
        return np.diag(self.nodes)/(2*k)


@register
class FlexJumper(Jumper):
    """Diagonal coefficients allowing order jump while still maintaining high stability"""

    aliases = ["FLEX-JUMPER", "FB2"]

    def computeQDelta(self, k=1):
        k = check(k)
        divider = 1 if k == 1 else 2*(k-1)
        return np.diag(self.nodes)/divider



import pytest
import numpy as np

from qmat.sdc import solveDahlquistSDC
from qmat.qcoeff.collocation import Collocation
from qmat import QDELTA_GENERATORS


@pytest.mark.parametrize("nNodes", [2, 3, 4])
@pytest.mark.parametrize("qDelta", ["BE", "FE"])
def testSweeps(qDelta, nNodes):
    
    coll = Collocation(nNodes=nNodes, nodeType="LEGENDRE", quadType="RADAU-RIGHT")
    gen = QDELTA_GENERATORS[qDelta](nodes=coll.nodes)

    runParams = dict(
        lam=1j, u0=1, T=np.pi, nSteps=10, nSweeps=nNodes,
        Q=coll.Q, 
    )

    QD1 = gen.getQDelta()
    uNum1 = solveDahlquistSDC(**runParams, QDelta=QD1)

    QD2 = gen.genCoeffs(k=[i+1 for i in range(nNodes)])
    uNum2 = solveDahlquistSDC(**runParams, QDelta=QD2)

    assert np.allclose(uNum1, uNum2), "solutions with 2D and 3D QDelta matrices are not the same"


@pytest.mark.parametrize("nNodes", [4, 6])
@pytest.mark.parametrize("nSteps", [10, 20])
@pytest.mark.parametrize("nSweeps", [2, 3, 4])
def testMonitors(nSweeps, nSteps, nNodes):
    coll = Collocation(nNodes=nNodes, nodeType="LEGENDRE", quadType="RADAU-RIGHT")
    gen = QDELTA_GENERATORS["BE"](nodes=coll.nodes)

    runParams = dict(
        lam=1j, u0=1, T=np.pi, nSteps=nSteps, nSweeps=nSweeps,
        Q=coll.Q, QDelta=gen.getQDelta(),
    )
    
    uNum = solveDahlquistSDC(**runParams)
    uNum2, monitors = solveDahlquistSDC(**runParams, monitors=["errors", "residuals"])
    
    assert np.allclose(uNum, uNum2), "solution with and without monitors are not the same"

    for key in ["errors", "residuals"]:
        assert key in monitors, f"'{key}' not in monitors"
        values = monitors[key]
        
        assert values.shape == (nSweeps+1, nSteps, nNodes), f"inconsistent shape for '{key}' values"
        assert np.all(np.abs(values[-1]) < np.abs(values[-2])), f"no decreasing {key}"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Defines the base abstract class to generate :math:`Q`-coefficients (Butcher tables) :
the :class:`QGenerator` 🚀

Each submodule contains specializations of this class for many kind of
methods :

- :class:`collocation` : Collocation based
- :class:`butcher` : Runge-Kutta based (Butcher tables)
"""
import numpy as np

from qmat.utils import checkOverriding, storeClass, importAll
from qmat.lagrange import LagrangeApproximation

class QGenerator(object):
    """Base abstract class for all :math:`Q`-coefficients generators"""

    @classmethod
    def getInstance(cls):
        """Provide an instance of this QGenerator using default parameters."""
        try:
            return cls()
        except TypeError:
            return cls(**cls.DEFAULT_PARAMS)

    @property
    def nodes(self):
        r"""Nodes :math:`\tau` (:math:`c` coefficients in Butcher table)"""
        raise NotImplementedError("mouahahah")

    @property
    def Q(self):
        r""":math:`Q` coefficients (:math:`A` Butcher table)"""
        raise NotImplementedError("mouahahah")

    @property
    def weights(self):
        r"""Weights :math:`\omega` (:math:`b` coefficients in Butcher table)"""
        raise NotImplementedError("mouahahah")

    @property
    def weightsEmbedded(self):
        """Weights for a secondary lower order method from the same stages."""
        raise NotImplementedError(f"no embedded weights implemented for {type(self).__name__}")

    @property
    def nNodes(self)->int:
        """Number of nodes (or stages) for this QGenerator"""
        return self.nodes.size

    @property
    def rightIsNode(self)->bool:
        """Wether or not the last nodes is the right boundary"""
        return self.nodes[-1] == 1.

    @property
    def T(self)->np.ndarray:
        """Transfer matrix from zero-to-nodes to node-to-node"""
        M = self.Q.shape[0]
        T = np.eye(M)
        T[1:,:-1][np.diag_indices(M-1)] = -1
        return T

    @property
    def S(self)->np.ndarray:
        """Quadrature matrix in node to node (N2N)"""
        Q = np.asarray(self.Q)
        M = self.Q.shape[0]
        T = np.eye(M)
        T[1:,:-1][np.diag_indices(M-1)] = -1
        return T @ Q

    @property
    def Tinv(self)->np.ndarray:
        """Transfer matrix from node-to-node to zero-to-node"""
        M = self.Q.shape[0]
        return np.tri(M)

    @property
    def hCoeffs(self)->np.ndarray:
        """:math:`h` interpolation coefficients for the right boundary"""
        approx = LagrangeApproximation(self.nodes)
        return approx.getInterpolationMatrix([1]).ravel()

    def genCoeffs(self, form="Z2N", hCoeffs=False, embedded=False):
        """
        Generate :math:`Q`-coefficients of this :class:`QGenerator` object.

        Parameters
        ----------
        form : str, optional
            Write coefficients in zero-to-nodes (Z2N) or node-to-node (N2N).
            The default is "Z2N".
        hCoeffs : bool, optional
            Wether or not returning the :math:`h` coefficients. The default is False.
        embedded : bool, optional
            Wether or not returning the embedded :math:`h` coefficients.
            The default is False.

        Returns
        -------
        out : tuple
            Contains (nodes, weights, Q).
            If `hCoeffs=True`, returns (nodes, weights, Q, hCoeffs).
            If `embedded=True`, `weights` is a 2xM array containing embedded weights in `weights[1]`.
        """
        if form == "Z2N":
            mat = self.Q
        elif form == "N2N":
            mat = self.S
        else:
            raise ValueError(f"form must be Z2N or N2N, not {form}")
        out = [self.nodes, self.weights, mat]
        if embedded:
            out[1] = np.vstack([out[1], self.weightsEmbedded])
        if hCoeffs:
            out.append(self.hCoeffs)
        return out

    @property
    def order(self):
        """Global convergence order of the method"""
        raise NotImplementedError("mouahahah")

    @property
    def orderEmbedded(self)->int:
        """Global convergence order of the associated embedded method"""
        return self.order - 1

    def solveDahlquist(self, lam, u0, T, nSteps, useEmbeddedWeights=False):
        r"""
        Solve the Dahlquist test problem

        .. math::

            \frac{du}{dt} = \lambda u, \quad t \in [0, T], \quad u(0)=u_0

        Parameters
        ----------
        lam : complex or float
            The :math:`\lambda` coefficient.
        u0 : complex or float
            The initial solution :math:`u_0`.
        T : float
            Final time :math:`T`.
        nSteps : int
            Number of time-step for the whole :math:`[0,T]` interval.
        useEmbeddedWeights : bool, optional
            Wether or not use the embedded weights for the prolongation. The default is False.

        Returns
        -------
        uNum : np.ndarray
            Array containing the `nSteps+1` solutions :math:`\{u(0), ..., u(T)\}`.
        """
        nodes, weights, Q = self.nodes, self.weights, self.Q

        if useEmbeddedWeights:
            weights = self.weightsEmbedded

        uNum = np.zeros(nSteps+1, dtype=complex)
        uNum[0] = u0

        dt = T/nSteps
        A = np.eye(nodes.size) - lam*dt*Q
        for i in range(nSteps):
            b = np.ones(nodes.size)*uNum[i]
            uStages = np.linalg.solve(A, b)
            uNum[i+1] = uNum[i] + lam*dt*weights.dot(uStages)

        return uNum

    def errorDahlquist(self, lam, u0, T, nSteps, uNum=None, useEmbeddedWeights=False):
        """
        Compute :math:`L_\infty` error in time for the Dahlquist problem

        Parameters
        ----------
        lam : complex or float
            The :math:`\lambda` coefficient.
        u0 : complex or float
            The initial solution :math:`u_0`.
        T : float
            Final time :math:`T`.
        nSteps : int
            Number of time-step for the whole :math:`[0,T]` interval.
        uNum : np.ndarray, optional
            Numerical solution, if not provided use the `solveDahlquist` method
            to compute the solution. The default is None.
        useEmbeddedWeights : bool, optional
            Wether or not use the embedded weights for the prolongation. The default is False.

        Returns
        -------
        float
            The :math:`L_\infty` norm.
        """
        if uNum is None:
            uNum = self.solveDahlquist(lam, u0, T, nSteps, useEmbeddedWeights=useEmbeddedWeights)
        times = np.linspace(0, T, nSteps+1)
        uExact = u0 * np.exp(lam*times)
        return np.linalg.norm(uNum-uExact, ord=np.inf)


Q_GENERATORS = {}
"""Dictionary containing all specialized :class:`QGenerator` classes, with all their aliases"""

def register(cls:QGenerator)->QGenerator:
    """Class decorator to register a specialized :class:`QGenerator` class in qmat"""
    # Check for correct overriding
    for name in ["nodes", "Q", "weights", "order"]:
        checkOverriding(cls, name)
    # Check that TEST_PARAMS are given and valid if no default constructor
    try:
        cls()
    except TypeError:
        try:
            params = cls.DEFAULT_PARAMS
        except AttributeError:
            raise AttributeError(
                f"{cls.__name__} requires DEFAULT_PARAMS attribute"
                " since it has no default constructor")
        try:
            cls(**params)
        except:
            raise TypeError(
                f"{cls.__name__} could not be instantiated with DEFAULT_PARAMS")
    # Store class (and aliases)
    storeClass(cls, Q_GENERATORS)
    return cls

def genQCoeffs(qType, form="Z2N", hCoeffs=False, embedded=False, **params):
    """
    Generate :math:`Q`-coefficients for a given method

    Parameters
    ----------
    qType : str
        Name (or alias) of the QGenerator.
    form : str, optional
        Write coefficients in zero-to-nodes (Z2N) or node-to-node (N2N).
        The default is "Z2N".
    hCoeffs : bool, optional
        Wether or not returning the :math:`h` coefficients. The default is False.
    embedded : bool, optional
        Wether or not returning the embedded :math:`h` coefficients.
        The default is False.
    **params :
        Parameters to be used to instantiate the QGenerator.

    Returns
    -------
    out : tuple
        Contains (nodes, weights, Q).
        If `hCoeffs=True`, returns (nodes, weights, Q, hCoeffs).
        If `embedded=True`, `weights` is a 2xM array containing embedded weights in `weights[1]`.
    """
    try:
        Generator = Q_GENERATORS[qType]
    except KeyError:
        raise ValueError(f"{qType=!r} is not available")
    gen = Generator(**params)
    return gen.genCoeffs(form, hCoeffs, embedded)


# Import all local submodules
__all__ = ["genQCoeffs", "QGenerator", "Q_GENERATORS", "register"]
importAll(locals(), __all__, __path__, __name__, __import__)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Utility function for `qmat`
"""
import inspect
import pkgutil

def checkOverriding(cls, name, isProperty=True):
    """Check if a class overrides a method with a given name"""
    method = getattr(cls, name)
    parent = getattr(cls.mro()[-2], name)
    assert method != parent, \
        f"{name} method must be overriden in {cls.__name__} class"
    if isProperty:
        assert type(method) == property, \
            f"{name} method must be a property in {cls.__name__} class"
    else:
        pass
        # TODO : check that signatures are the same


def checkGenericConstr(cls):
    """Check if a class implement a constructor with a `**kwargs` generic parameter"""
    sig = inspect.signature(cls.__init__)
    try:
        par = sig.parameters["kwargs"]
        assert par.kind == par.VAR_KEYWORD
    except (KeyError, AssertionError):
        raise AssertionError(f"{cls.__name__} class requires **kwargs in its constructor")


def storeAlias(cls, dico, alias):
    """Store a class into a dictionary with a given alias"""
    assert alias not in dico, f"{alias} alias already registered in {dico}"
    dico[alias] = cls


def storeClass(cls, dico):
    """Store a class into a dictionary"""
    storeAlias(cls, dico, cls.__name__)
    aliases = getattr(cls, "aliases", None)
    if aliases:
        assert isinstance(aliases, list), \
            f"aliases must be a list in class {cls.__name__}"
        for alias in aliases:
            storeAlias(cls, dico, alias)


def importAll(localVars, __all__, __path__, __name__, __import__):
    """Import all submodules in the current (sub-)package"""
    __all__ += [var for var in localVars.keys() if not var.startswith('__')]
    for _, moduleName, _ in pkgutil.walk_packages(__path__):
        __all__.append(moduleName)
        __import__(__name__+'.'+moduleName)


def getClasses(dico, module=None):
    """Retrieve all classes stored into a dictionary, filtering aliases"""
    classes = {}
    if module is None:
        check = lambda cls: True
    else:
        check = lambda cls: cls.__module__.endswith("."+module)
    for key, cls in dico.items():
        if cls not in classes.values() and check(cls):
            classes[key] = cls
    return classes



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Submodule for QDelta coefficients based on time-stepping scheme.
Allows to build equivalent SDC sweeps as those introduced in the original
paper from `[Dutt, Greengard & Rokhlin, 2000] <https://link.springer.com/article/10.1023/A:1022338906936>`_.

Examples
--------
>>> from qmat.qcoeff.collocation import Collocation
>>> coll = Collocation(nNodes=4, nodeType="LEGENDRE", quadType="RADAU-RIGHT")
>>>
>>> from qmat import genQDeltaCoeffs
>>> QDelta = genQDeltaCoeffs("IE", nodes=coll.nodes)
>>>
>>> from qmat.qdelta.timestepping import TRAP
>>> gen = TRAP(nodes=coll.nodes)
>>> SDelta, dTau = gen.genCoeffs(form="N2N", dTau=True)
"""
import numpy as np

from qmat.qdelta import QDeltaGenerator, register


class TimeStepping(QDeltaGenerator):
    """
    Base class for time-stepping based :math:`Q_\Delta` approximations

    Parameters
    ----------
    nodes : list-like
        Normalized nodes in increasing order.
    tLeft : float, optional
        Left bound for the nodes. The default is 0.
    **kwargs :
        Additional parameters given in a generic call, ignored by this class.
    """
    def __init__(self, nodes, tLeft=0, **kwargs):
        nodes = np.asarray(nodes)
        deltas = nodes.copy()
        deltas[0] = nodes[0] - tLeft
        deltas[1:] = np.ediff1d(nodes)

        self.deltas:np.ndarray = deltas
        """Differences between nodes"""

        self.nodes:np.ndarray = nodes
        """Array of normalized nodes"""

        self.tLeft:float = tLeft
        """Left bound for the nodes"""

    @property
    def size(self)->int:
        return self.nodes.size


@register
class BE(TimeStepping):
    """Approximation based on Backward Euler steps between the nodes"""
    aliases = ["IE"]

    def computeQDelta(self, k=None):
        QDelta, M, deltas = self.zeros, self.nodes.size, self.deltas
        for i in range(M):
            QDelta[i:, :M-i] += np.diag(deltas[:M-i])
        return QDelta


@register
class FE(TimeStepping):
    """Approximation based on Forward Euler steps between the nodes"""
    aliases = ["EE"]

    def computeQDelta(self, k=None):
        QDelta, M, deltas = self.zeros, self.nodes.size, self.deltas
        for i in range(1, M):
            QDelta[i:, :M-i] += np.diag(deltas[1:M-i+1])
        return QDelta

    @property
    def dTau(self)->np.ndarray:
        return self.nodes*0 + self.deltas[0]


@register
class TRAP(TimeStepping):
    """Approximation based on Trapezoidal Rule between the nodes"""
    aliases = ["CN"]

    def computeQDelta(self, k=None):
        QDelta, M, deltas = self.zeros, self.nodes.size, self.deltas
        for i in range(0, M):
            QDelta[i:, :M-i] += np.diag(deltas[:M-i])
        for i in range(1, M):
            QDelta[i:, :M-i] += np.diag(deltas[1:M-i+1])
        QDelta /= 2.0
        return QDelta

    @property
    def dTau(self)->np.ndarray:
        return self.nodes*0 + self.deltas[0]/2.0


@register
class BEPAR(TimeStepping):
    """Approximation based on parallel Backward Euler steps from zero to nodes"""
    aliases = ["IEpar"]

    def computeQDelta(self, k=None):
        return np.diag(self.nodes) - self.tLeft


@register
class TRAPAR(TimeStepping):
    """Approximation based on parallel Trapezoidal Rule from zero to nodes"""

    def computeQDelta(self, k=None):
        return np.diag(self.nodes/2) - self.tLeft

    @property
    def dTau(self)->np.ndarray:
        return self.nodes/2.0 - self.tLeft



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Defines the base abstract class to generate  :math:`Q_\Delta` approximations :
the :class:`QDeltaGenerator` 🚀

Each submodule contains specializations of this class for many kind of
methods :

- :class:`timestepping` : based on time-stepping methods (Backward Euler, etc ...)
- :class:`algebraic` : based on algebraic consideration on the :math:`Q` matrix
- :class:`min` : diagonal approximations based on minimization
"""
import inspect
import numpy as np

from qmat.utils import checkOverriding, storeClass, importAll, checkGenericConstr


class QDeltaGenerator(object):
    r"""
    Base abstract class for :math:`Q_\Delta` coefficients generators.

    Parameters
    ----------
    Q : np.ndarray
        The :math:`Q` matrix of the base approximated method.
    **kwargs :
        Additional parameters given in a generic call, ignored by the class.
    """

    _K_DEPENDENT = False
    """Wether or not the :math:`Q_\Delta` coefficients varies with the iterations"""

    def __init__(self, Q, **kwargs):
        self.Q:np.ndarray = np.asarray(Q, dtype=float)
        """:math:`Q` matrix of the approximated time-integration method"""

    @property
    def size(self)->int:
        """Dimension of the approximated :math:`Q`-coefficients (number of nodes)"""
        return self.Q.shape[0]

    @property
    def zeros(self)->np.ndarray:
        """Seros matrix with the same size of the underlying :math:`Q` matrix"""
        M = self.size
        return np.zeros((M, M), dtype=float)

    def computeQDelta(self, k=None) -> np.ndarray:
        """
        Compute and returns the :math:`Q_\Delta` matrix, has to be implemented
        in the specialized class.

        Parameters
        ----------
        k : int, optional
            Iteration number of the approximation. The default is None.

        Returns
        -------
        QDelta : np.ndarray
        """
        raise NotImplementedError("mouahahah")

    def getQDelta(self, k=None, copy=True):
        r"""
        Generic method to retrieve the :math:`Q_\Delta` coefficients

        Parameters
        ----------
        k : int, optional
            Iteration number of the approximation (if needed). The default is None.
        copy : bool, optional
            Return a copy of the the result returned by `computeQDelta`.
            The default is True.

        Returns
        -------
        QDelta : np.ndarray
        """
        try:
            QDelta = self._QDelta[k] if self._K_DEPENDENT else self._QDelta
        except Exception as e:
            QDelta = self.computeQDelta(k)
            if type(e) == AttributeError:
                self._QDelta = {k: QDelta} if self._K_DEPENDENT else QDelta
            elif type(e) == KeyError:
                self._QDelta[k] = QDelta
            else:
                raise Exception("some very weird bug happened ... did you do fishy stuff ?")
        return QDelta.copy() if copy else QDelta

    def getSDelta(self, k=None):
        r"""
        Compute the :math:`S_\Delta` matrix (approximation of :math:`S`).

        Parameters
        ----------
        k : int, optional
            Iteration number, used when the approximation depends on it.
            The default is None.

        Returns
        -------
        SDelta : np.ndarray
        """
        QDelta = self.getQDelta(k)
        M = QDelta.shape[0]
        T = np.eye(M)
        T[1:,:-1][np.diag_indices(M-1)] = -1
        return T @ QDelta

    @property
    def dTau(self)->np.ndarray:
        r"""The :math:`\delta_\tau` coefficients associated to :math:`Q_\Delta`"""
        return np.zeros(self.size, dtype=float)

    def genCoeffs(self, k=None, form="Z2N", dTau=False):
        r"""
        Generic method to produce :math:`Q_\Delta` coefficients

        Parameters
        ----------
        k : int or list, optional
            Iteration(s) for the approximation. The default is None.
        form : str, optional
            Build approximation in zero-to-nodes (Z2N) or node-to-node (N2N).
            The default is "Z2N".
        dTau : bool, optional
            Wether or not to return the :math:`\delta_\tau`.
            The default is False.

        Returns
        -------
        np.ndarray or tuple
            If `k` is a scalar or `None`, returns a MxM matrix.
            If `k` is a list, returns a len(k)xMxM matrix.
            If `dTau=True`, returns a tuple `(QDelta, dTau)`.
        """
        if form == "Z2N":
            gen = lambda k, copy=False: self.getQDelta(k, copy)
        elif form == "N2N":
            gen = lambda k, copy=None: self.getSDelta(k)
        else:
            raise ValueError(f"form must be Z2N or N2N, not {form}")
        try:
            k = list(k)
            out = [np.array([gen(_k, copy=False) for _k in k])]
        except TypeError:
            out = [gen(k)]
        if dTau:
            out += [self.dTau]
        return out if len(out) > 1 else out[0]


QDELTA_GENERATORS = {}
"""Dictionary containing all specialized :class:`QDeltaGenerator` classes"""

def register(cls:QDeltaGenerator)->QDeltaGenerator:
    """Class decorator to register a specialized :class:`QDeltaGenerator` class in `qmat`"""
    checkGenericConstr(cls)
    checkOverriding(cls, "computeQDelta", isProperty=False)
    try:
        sig = inspect.signature(cls.computeQDelta)
        par = sig.parameters["k"]
        assert par.kind == par.POSITIONAL_OR_KEYWORD
        if par.default is not None:
            cls._K_DEPENDENT = True
    except (KeyError, AssertionError):
        raise AssertionError(f"{cls.__name__} class does not properly override the computeQDelta method")
    storeClass(cls, QDELTA_GENERATORS)
    return cls


def genQDeltaCoeffs(qDeltaType, nSweeps=None, form="Z2N", dTau=False, **params):
    r"""
    Generic function to produce :math:`Q_\Delta` coefficients

    Parameters
    ----------
    qDeltaType : str or list
        The type of approximation, can be a list to have several sweeps.
    nSweeps : int, optional
        Number of sweeps when :math:`Q_\Delta` matrices are required for
        several sweeps. The default is None.
    form : str, optional
        Build approximation in zero-to-nodes (Z2N) or node-to-node (N2N).
        The default is "Z2N".
    dTau : bool, optional
        Wether or not to return the :math:`\delta_\tau`. The default is False.
    **params
        Additional arguments used to instantiate all :class:`QDeltaGenerator`

    Returns
    -------
    np.ndarray or tuple
        If `qDeltaType` is a string, returns a :math:`M \times M` matrix.
        If `qDeltaType` is a list or `nSweeps != None`,
        returns a :math:`N_{sweeps} \times M \times M` matrix.
        If `dTau=True`, returns a tuple `(QDelta, dTau)`.
    """
    # Check arguments
    if isinstance(qDeltaType, str):
        if nSweeps is None:
            pass  # only one QDelta matrix, default approach
        elif isinstance(nSweeps, int) and nSweeps > 0:
            qDeltaType = [qDeltaType]  # more sweeps of the same QDelta matrix
        else:
            raise ValueError(f"bad value of nSweep {nSweeps}")
    elif isinstance(qDeltaType, list):
        assert len(qDeltaType) > 0, "need at least one qDeltaType in the list"
        if nSweeps is None:
            nSweeps = len(qDeltaType)  # number of sweeps given in the list
        elif isinstance(nSweeps, int) and nSweeps > 0:
            # complete with additional sweeps
            assert nSweeps >= len(qDeltaType), \
                f"nSweeps ({nSweeps}) is lower than list length for qDeltaType ({qDeltaType})"
            qDeltaType += [qDeltaType[-1]]*(nSweeps-len(qDeltaType))
        else:
            raise ValueError(f"bad value of nSweep {nSweeps}")
    else:
        raise ValueError(f"bad value of qDeltaType {qDeltaType}")

    if nSweeps is None:  # Single matrix return

        try:
            Generator = QDELTA_GENERATORS[qDeltaType]
        except KeyError:
            raise ValueError(f"qDeltaType={qDeltaType} is not available")

        gen = Generator(**params)
        return gen.genCoeffs(form=form, dTau=dTau)

    else:  # Multiple matrices return
        try:
            Generators = [QDELTA_GENERATORS[qDT] for qDT in qDeltaType]
        except KeyError:
            raise ValueError(f"qDeltaType={qDeltaType} is not available")

        if len(qDeltaType) == 1:  # Single QDelta generator
            gen = Generators[0](**params)
            return gen.genCoeffs(
                k=[k+1 for k in range(nSweeps)], form=form, dTau=dTau)

        else:  # Multiple QDelta generators
            gens = [Gen(**params) for Gen in Generators]
            out = [np.array(
                [gen.genCoeffs(k+1, form) for k, gen in enumerate(gens)]
                )]
            if dTau:
                out += [gens[0].dTau]

    return out if len(out) > 1 else out[0]


# Import all local submodules
__all__ = ["genQDeltaCoeffs", "QDeltaGenerator", "QDELTA_GENERATORS", "register"]
importAll(locals(), __all__, __path__, __name__, __import__)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
r"""
**Main sub-packages** 📦

- :class:`qcoeff` : to generate the :math:`Q`-coefficients (Butcher tables)
- :class:`qdelta` : to generate :math:`Q_\Delta` approximations for :math:`Q` matrices

**Utility modules** ⚙️

- :class:`lagrange` : Barycentric polynomial approximations (integral, interpolation, derivation)
- :class:`nodes` : generation of multiple types of quadrature nodes
- :class:`sdc` : utility function to run SDC on simple problems
- :class:`mathutils` : utility functions for math operations
- :class:`utils` : utility functions for the whole package

Examples
--------
>>> from qmat import genQCoeffs, genQDeltaCoeffs
>>>
>>> # Coefficients of a specific collocation method
>>> nodes, weights, Q = genQCoeffs(
>>>     "Collocation", nNodes=4, nodeType="LEGENDRE", quadType="RADAU-RIGHT")
>>>
>>> # QDelta matrix from Implicit-Euler based SDC
>>> QDelta = genQDeltaCoeffs("IE", nodes=nodes)
>>>
>>> # Butcher table of the classical explicit RK4 method
>>> c, b, A = genQCoeffs("ERK4")

>>> from qmat import Q_GENERATORS, QDELTA_GENERATORS
>>> print(Q_GENERATORS)         # list all available generator classes for Q coefficients
>>> print(QDELTA_GENERATORS)    # list all available generator classes for QDelta approximations
"""
from qmat.qcoeff import genQCoeffs, Q_GENERATORS
from qmat.qdelta import genQDeltaCoeffs, QDELTA_GENERATORS

__all__ = [
    "genQCoeffs", "genQDeltaCoeffs",
    "Q_GENERATORS", "QDELTA_GENERATORS"]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Base module for math utility functions
"""
import numpy as np


def numericalOrder(nSteps, err):
    """
    Compute numerical order from two vectors containing the error and the number of time-steps.

    Parameters
    ----------
    nSteps : np.1darray or list
        Different number of steps to compute the error.
    err : np.1darray
        Diffenrent error values associated to the number of steps.

    Returns
    -------
    beta : float
        Order coefficient computed through linear regression.
    rmse : float
        The root mean square error of the linear regression.
    """
    nSteps = np.asarray(nSteps)
    x, y = np.log10(1/nSteps), np.log10(err)

    # Compute regression coefficients and rmse
    xMean = x.mean()
    yMean = y.mean()
    sX = ((x-xMean)**2).sum()
    sXY = ((x-xMean)*(y-yMean)).sum()

    beta = sXY/sX
    alpha = yMean - beta*xMean

    yHat = alpha + beta*x
    rmse = ((y-yHat)**2).sum()**0.5
    rmse /= x.size**0.5

    return beta, rmse


def lduFactorization(A:np.ndarray):
    """
    Perform LDU factorization on a square matrix A.
    
    Parameters
    ----------
    A : np.2darray 
        The square matrix to factorize (n x n).
    
    Returns
    -------
    L : np.2darray
        Lower triangular matrix with ones on the diagonal.
    D : np.2darray 
        Diagonal matrix.
    U : np.2darray 
        Upper triangular matrix with ones on the diagonal.
    """
    # Ensure A is a square matrix
    n, m = A.shape
    assert n == m, "Matrix A must be square."

    # Initialize L, D, U matrices
    L = np.eye(n)
    D = np.zeros((n, n))
    U = np.eye(n)

    # Decompose A into L, D, U
    for i in range(n):
        # Compute D[i, i] as the diagonal element
        D[i, i] = A[i, i] - np.sum(L[i, :i] * D[:i, :i].diagonal() * U[:i, i])

        for j in range(i + 1, n):
            # Compute elements for L below the diagonal
            L[j, i] = (A[j, i] - np.sum(L[j, :i] * D[:i, :i].diagonal() * U[:i, i])) / D[i, i]
            
            # Compute elements for U above the diagonal
            U[i, j] = (A[i, j] - np.sum(L[i, :i] * D[:i, :i].diagonal() * U[:i, j])) / D[i, i]

    return L, D, U


def getExtrapolationMatrix(nodes, times, pOrder=None):
    """
    Generate a polynomial based extrapolation matrix,
    base on polynomial regression of a function represented
    on given nodes.

    Parameters
    ----------
    nodes : np.1darray like, shape (M,)
        The nodes where function values are known.
    times : np.1darray like, shape (N,)
        The times where to extrapolate the polynomial.
    pOrder : int
        Order of the polynomial regression on the node values
        (default = len(nodes)-1).

    Returns
    -------
    P : np.2darray, shape (N, M)
        Extrapolation matrix, that can be used on any node values.
    """
    if pOrder is None: pOrder = np.size(nodes)-1
    X = np.vander(nodes, N=pOrder+1, increasing=True)
    T = np.vander(times, N=pOrder+1, increasing=True)
    P = T @ np.linalg.solve(X.T @ X, X.T)
    return P


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Functions to run SDC and evaluate its numerical error on simple problems.
"""
import numpy as np


def solveDahlquistSDC(lam, u0, T, nSteps:int, nSweeps:int, Q:np.ndarray, QDelta:np.ndarray,
                      weights=None, monitors=None):
    r"""
    Solve the Dahlquist problem with SDC.

    Parameters
    ----------
    lam : complex or float
        The :math:`\lambda` coefficient.
    u0 : complex or float
        The initial solution :math:`u_0`.
    T : float
        Final time :math:`T`.
    nSteps : int
        Number of time-step for the whole :math:`[0,T]` interval.
    nSweeps : int
        Number of SDC sweeps.
    Q : np.ndarray
        Quadrature matrix :math:`Q` used for SDC.
    QDelta : np.ndarray
        Approximate quadrature matrix :math:`Q_\Delta` used for SDC. 
        If three dimensional, use the first dimension for the sweep index.
    weights : np.ndarray, optional
        Quadrature weights to use for the prologation.
        If None, prolongation is not performed. The default is None.

    Returns
    -------
    uNum : np.ndarray
        Array containing the `nSteps+1` solutions :math:`\{u(0), ..., u(T)\}`.
    """
    nodes = Q.sum(axis=1)
    nNodes = Q.shape[0]
    dt = T/nSteps
    times = np.linspace(0, T, nSteps+1)
    
    QDelta = np.asarray(QDelta)
    if QDelta.ndim == 3:
        assert QDelta.shape == (nSweeps, nNodes, nNodes), "inconsistent shape for QDelta"
    else:
        assert QDelta.shape == (nNodes, nNodes), "inconsistent shape for QDelta"
        QDelta = np.repeat(QDelta[None, ...], nSweeps, axis=0)

    # Preconditioner built for each sweeps
    P = np.eye(nNodes)[None, ...] - lam*dt*QDelta

    uNum = np.zeros(nSteps+1, dtype=complex)
    uNum[0] = u0

    # Setup monitors if any
    if monitors:
        tmp = {}
        for key in monitors:
            assert key in ["residuals", "errors"], f"unknown key '{key}' for monitors"
            tmp[key] = np.zeros((nSweeps+1, nSteps, nNodes), dtype=complex)
        monitors = tmp

    for i in range(nSteps):

        uNodes = np.ones(nNodes)*uNum[i]

        # Monitoring
        if monitors:
            if "residuals" in monitors:
                monitors["residuals"][0, i] = uNum[i] + lam*dt*Q @ uNodes - uNodes
            if "errors" in monitors:
                monitors["errors"][0, i] = uNodes - u0*np.exp(lam*(times[i] + dt*nodes))

        for k in range(nSweeps):
            b = uNum[i] + lam*dt*(Q-QDelta[k]) @ uNodes
            uNodes = np.linalg.solve(P[k], b)

            # Monitoring
            if monitors:
                if "residuals" in monitors:
                    monitors["residuals"][k+1, i] = uNum[i] + lam*dt*Q @ uNodes - uNodes
                if "errors" in monitors:
                    monitors["errors"][k+1, i] = uNodes - u0*np.exp(lam*(times[i] + dt*nodes))

        if weights is None:
            uNum[i+1] = uNodes[-1]
        else:
            uNum[i+1] = uNum[i] + lam*dt*weights.dot(uNodes)

    if monitors:
        return uNum, monitors
    else:   
        return uNum


def errorDahlquistSDC(lam, u0, T, nSteps, nSweeps, Q, QDelta,
                      weights=None, uNum=None):
    r"""
    Compute the time :math:`L_\infty` error of SDC.

    Parameters
    ----------
    lam : complex or float
        The :math:`\lambda` coefficient.
    u0 : complex or float
        The initial solution :math:`u_0`.
    T : float
        Final time :math:`T`.
    nSteps : int
        Number of time-step for the whole :math:`[0,T]` interval.
    nSweeps : int
        Number of SDC sweeps.
    Q : np.ndarray
        Quadrature matrix :math:`Q` used for SDC.
    QDelta : np.ndarray
        Approximate quadrature matrix :math:`Q_\Delta` used for SDC.
    weights : np.ndarray, optional
        Quadrature weights to use for the prologation.
        If None, prolongation is not performed. The default is None.
    uNum : np.ndarray, optional
        Numerical solution, if not provided use the `solveDahlquist` method
        to compute the solution. The default is None.

    Returns
    -------
    float
        The :math:`L_\infty` norm.
    """
    if uNum is None:
        uNum = solveDahlquistSDC(
            lam, u0, T, nSteps, nSweeps, Q, QDelta,
            weights=weights)

    times = np.linspace(0, T, nSteps+1)
    uExact = u0 * np.exp(lam*times)
    return np.linalg.norm(uNum-uExact, ord=np.inf)


def getOrderSDC(coll, nSweeps, qDelta, prolongation):
    r"""
    Give the expected order of SDC after a fixed number of iterations.

    Parameters
    ----------
    coll : :class:`qmat.qcoeff.collocation.Collocation`
        The underlying `Collocation` class.
    nSweeps : int
        Number of sweeps for SDC.
    qDelta : str
        Type of the :math:`Q_\Delta` approximation used.
    prolongation : bool
        Wether or not the prolongation is done at the end.

    Returns
    -------
    order : int
        Expected order of the SDC time-integration.
    """
    # TODO : extend with additional results from
    # https://gitlab.inria.fr/sweet/sweet/-/blob/main/mule_local/python/sdc/qmatrix.py#L596

    nNodes, nodeType, quadType = coll.nodes.size, coll.nodeType, coll.quadType

    # Maximum order from the collocation problem
    maxOrder = coll.order

    # Order of SDC
    order = 0
    if nSweeps > 0:
        # first sweep
        if qDelta in ['TRAPAR', 'TRAP']:
            order += 2
        else:
            order += 1
        # rest of sweeps
        order += nSweeps-1
    # take into account prolongation
    if prolongation == "QUADRATURE":
        order += 1

    order = min(maxOrder, order)

    # Edge cases with bonus order (or malus some times ...)
    # TODO: couple with the Butcher theory from Joscha to retrieve this theoretically ...
    if prolongation == "QUADRATURE":  # COPY initialization
        if qDelta == "TRAP":
            if nSweeps == 1 and nNodes == 3 and nodeType == "EQUID" and quadType == "RADAU-LEFT":
                order += 1
            if nSweeps == 2 and nNodes == 4 and nodeType == "CHEBY-2" and quadType == "LOBATTO":
                order += 1
            if nSweeps == 2 and nNodes == 4 and nodeType == "CHEBY-3" and quadType == "RADAU-LEFT":
                order += 1
            if nSweeps == 2 and nNodes == 4 and nodeType == "CHEBY-4" and quadType in ["RADAU-RIGHT", "LOBATTO"]:
                order += 1
        if qDelta == "TRAPAR":
            if nSweeps == 2 and nNodes == 3 and nodeType == "CHEBY-3" and quadType == "RADAU-LEFT":
                order += 1
            if nSweeps == 2 and nNodes == 3 and nodeType == "CHEBY-4" and quadType == "RADAU-RIGHT":
                order += 1
            if nSweeps == 3 and nNodes == 3 and nodeType == "CHEBY-3" and quadType == "RADAU-LEFT":
                order += 1
        if qDelta == "BE":
            if nSweeps == 3 and nNodes == 3 and nodeType == "CHEBY-3" and quadType == "RADAU-LEFT":
                order += 1
            if nSweeps == 3 and nNodes == 3 and nodeType == "CHEBY-4" and quadType == "RADAU-RIGHT":
                order += 1
        if qDelta == "FE":
            if nSweeps == 3 and nNodes == 3 and nodeType == "CHEBY-1" and quadType == "RADAU-RIGHT":
                order += 1
            if nSweeps == 3 and nNodes == 3 and nodeType == "CHEBY-3" and quadType in ["RADAU-LEFT", "LOBATTO"]:
                order += 1
            if nSweeps == 3 and nNodes == 3 and nodeType == "CHEBY-4" and quadType == "RADAU-RIGHT":
                order += 1
        if qDelta == "BEPAR":
            if nSweeps == 3 and nNodes == 3 and nodeType == "EQUID" and quadType == "RADAU-LEFT":
                order += 1
            if nSweeps == 3 and nNodes == 3 and nodeType in ["CHEBY-1", "CHEBY-2"] and quadType in ["RADAU-LEFT", "RADAU-RIGHT"]:
                order += 1
            if nSweeps == 3 and nNodes == 3 and nodeType in ["CHEBY-3", "CHEBY-4"]:
                order += 1
        if qDelta == "TRAPAR":
            if nSweeps == 3 and nNodes == 4 and nodeType == "EQUID":
                order += 1
            if nSweeps == 3 and nNodes == 4 and nodeType in ["CHEBY-1", "CHEBY-2", "CHEBY-3", "CHEBY-4"]:
                order += 1

    if prolongation == "LASTNODE":
        if qDelta == "BE":
            if nSweeps == 4 and nNodes == 3 and nodeType == "CHEBY-4" and quadType == "RADAU-RIGHT":
                order += 1
        if qDelta == "FE":
            if nSweeps == 3 and nNodes == 3 and nodeType == "EQUID" and quadType == "RADAU-RIGHT":
                order += 1
            if nSweeps == 4 and nNodes == 3 and nodeType == "CHEBY-4" and quadType == "RADAU-RIGHT":
                order += 1
        if qDelta == "TRAP":
            if nSweeps == 2 and nNodes == 3 and nodeType in ["EQUID", "CHEBY-1", "CHEBY-2"] and quadType == "LOBATTO":
                order += 1
            if nSweeps == 2 and nNodes == 3 and nodeType == "LEGENDRE":
                order += 1
            if nSweeps == 2 and nNodes == 4 and nodeType == "EQUID" and quadType == "LOBATTO":
                order += 2
            if nSweeps == 2 and nNodes == 4 and nodeType == "EQUID" and quadType == "RADAU-RIGHT":
                order += 1
            if nSweeps == 2 and nNodes == 4 and nodeType in ["LEGENDRE", "CHEBY-1", "CHEBY-2", "CHEBY-3", "CHEBY-4"]:
                order += 1
            if nSweeps == 3 and nNodes == 3 and nodeType == "LEGENDRE" and quadType == "RADAU-RIGHT":
                order += 1
            if nSweeps == 4 and nNodes == 4 and nodeType == "LEGENDRE" and quadType == "LOBATTO":
                order += 1
            if nSweeps == 4 and nNodes == 4 and nodeType == "LEGENDRE" and quadType == "RADAU-RIGHT":
                order += 2
        if qDelta == "BEPAR":
            if nSweeps == 3 and nNodes == 2 and nodeType == "CHEBY-2" and quadType == "RADAU-RIGHT":
                order += 1
            if nSweeps == 4 and nNodes == 3 and nodeType in ["EQUID", "CHEBY-1", "CHEBY-2", "CHEBY-3", "CHEBY-4"] and quadType == "RADAU-RIGHT":
                order += 1
            if nSweeps == 4 and nNodes == 3 and nodeType in ["CHEBY-3", "CHEBY-4"] and quadType == "LOBATTO":
                order += 1
        if qDelta == "TRAPAR":
            if nSweeps == 3 and nNodes == 3 and nodeType == "CHEBY-4" and quadType == "RADAU-RIGHT":
                order += 1
            if nSweeps == 4 and nNodes == 4 and nodeType in ["EQUID", "CHEBY-1", "CHEBY-2", "CHEBY-3", "CHEBY-4"]:
                order += 1


    return order
